﻿using System;
using System.Drawing;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Linq;
using ESRI.ArcGIS.Analyst3D;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.ArcScene;
using ESRI.ArcGIS.Desktop;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.SystemUI;
using ESRI.ArcGIS.Geometry;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.DisplayUI;
//using ESRI.ArcGIS.ADF.BaseClasses;
using System.Windows.Forms;
using Microsoft.Kinect;
using Microsoft.Kinect.Toolkit.Interaction;
using System.Diagnostics;
using System.Runtime.InteropServices;
//using ESRI.ArcGIS.ADF.BaseClasses;
//using System.Threading;

namespace ArcSceneAddin2
{
    public struct Position
    {
        public double x;
        public double y;
        public double z;
    }

    public class Tool1 : ESRI.ArcGIS.Desktop.AddIns.Tool
    {
        int m_gestures;
        ICamera3 cam;
        ISceneGraph scene;
        private KinectSensor sensor;
        UserInfo[] userInfos;
        private InteractionStream interactionStream { get; set; }
        Skeleton m_closestSkeleton;
        Position m_lastPos;
        double m_zoom;
        double m_rotX;
        double m_rotY;
        double m_moveZStart;
        bool m_rightGrip;
        bool m_leftGrip;
        MyInteractionClient m_interactionClient;
        Skeleton[] skeletons;

        static WaveGesture m_waveLeftHand;
        static WaveGesture m_waveRightHand;
        bool dynamicGestures;

        public Tool1()
        {
            m_gestures = 0;
            userInfos = null;
            m_interactionClient = new MyInteractionClient();
            m_rightGrip = false;
            m_leftGrip = false;
            m_zoom = 0.0;
            m_rotX = 0.0;
            m_rotY = 0.0;
            m_moveZStart = 0.0;
            m_lastPos.x = 0.0;
            m_lastPos.y = 0.0;
            m_lastPos.z = 0.0;
            m_closestSkeleton = null;
            m_waveLeftHand = new WaveGesture(JointType.HandLeft, JointType.WristLeft, JointType.ElbowLeft, JointType.HandRight, JointType.ElbowRight);
            m_waveRightHand = new WaveGesture(JointType.HandRight, JointType.WristRight, JointType.ElbowRight, JointType.HandLeft, JointType.ElbowLeft);
            dynamicGestures = false;
        }

        /*public int Bitmap
        {
            get
            {
                return m_hBitmap.ToInt32();
            }
        }*/

        protected override void OnActivate()
        {
            base.OnActivate(); //zavolání metody předka
            cam = new CameraClass(); // nová instance kamery
            cam = (ICamera3)ArcScene.Scene.SceneGraph.CurrentViewer.Camera; // předaní kamery z aktuálního pohledu
            scene = ArcScene.Scene.SceneGraph; // předání celé scény ArcScene

            scene.SetupProjection(cam); // nastavení projekce

            StartSensor(); // spuštění zařízení Kinect
        }

        protected override bool OnDeactivate()
        {
            if (this.sensor != null)
                this.sensor.Stop();

            return base.OnDeactivate();
        }

        private void SensorSkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            if (interactionStream == null)
            {
                interactionStream = new InteractionStream(this.sensor, new MyInteractionClient());
                interactionStream.InteractionFrameReady += new EventHandler<InteractionFrameReadyEventArgs>(interactionStream_InteractionFrameReady);
            }

            skeletons = null; //pole do ktereho si nasypu vsechny dostupne skeletony (v danem frame)

            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null) //pokud jsou data k dispozici
                {
                    skeletons = new Skeleton[skeletonFrame.SkeletonArrayLength]; //inicializace pole dle poctu skeletonu
                    skeletonFrame.CopySkeletonDataTo(skeletons);                 //nakopiruji si do sveho pole vsechny skeletony

                    try
                    {
                        interactionStream.ProcessSkeleton(skeletons, this.sensor.AccelerometerGetCurrentReading(), skeletonFrame.Timestamp);
                        ProcessDepthFrame(skeletonFrame.Timestamp);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.Message);
                    }
                }
            }

            //vrati nejblizsiho skeletona
            m_closestSkeleton = skeletons.Where(s => s.TrackingState == SkeletonTrackingState.Tracked).
                                OrderBy(s => s.Position.Z /* Math.Abs(s.Position.X)*/).FirstOrDefault();

            //pokud nemam skeletona tak koncim
            if (m_closestSkeleton == null)
                return;

            if (CheckJointsForTracking(m_closestSkeleton))
            {
                ProcessGestures();
            }
        }

        private void ProcessGestures()
        {
            if (m_gestures == 0 || m_gestures == 1)
                moveXYZ();

            if (m_gestures == 0 || m_gestures == 2)
                zoom();

            if (m_gestures == 0 || m_gestures == 3)
                rotation();

            //if (m_gestures == 0)
                m_waveLeftHand.Update(m_closestSkeleton, m_leftGrip);

            //if (m_gestures == 0)
                m_waveRightHand.Update(m_closestSkeleton, m_rightGrip);
        }

        private void rotation()
        {
            var leftHand = m_closestSkeleton.Joints[JointType.HandLeft];
            var rightHand = m_closestSkeleton.Joints[JointType.HandRight];
            var leftElbow = m_closestSkeleton.Joints[JointType.ElbowLeft];
            var rightElbow = m_closestSkeleton.Joints[JointType.ElbowRight];

            bool left = false;
            bool right = false;

            if (leftHand.Position.Y > leftElbow.Position.Y && m_leftGrip)
                left = true;

            if (rightHand.Position.Y > rightElbow.Position.Y && m_rightGrip)
                right = true;
            
            if (!left || !right)
            {
                m_rotY = 0.0;
                m_rotX = 0.0;
                m_gestures = 0;
                return;
            }

            m_gestures = 3;

            if (m_rotY != 0.0)
            {
                double y = (double)rightHand.Position.Z - (double)leftHand.Position.Z;
                double x = (double)rightHand.Position.Y - (double)leftHand.Position.Y;
                double azimuth = 0.0;
                double inclin = 0.0;

                if (y > m_rotY + 0.007)
                {
                    if (dynamicGestures)
                        azimuth = y * 10;
                    else
                        azimuth = 5;
                }
                else
                    if (y < m_rotY - 0.007)
                    {
                        if (dynamicGestures)
                            azimuth = y * 10;
                        else
                            azimuth = -5;
                    }

                if (x > m_rotX + 0.01)
                {
                    if (dynamicGestures)
                        inclin = x * 10;
                    else
                        inclin = 5;
                }
                else
                    if (x < m_rotX - 0.01)
                    {
                        if (dynamicGestures)
                            inclin = x * 10;
                        else
                            inclin = -5;
                    }

                cam.PolarUpdate(1, azimuth, inclin, true);

                scene.CurrentViewer.Redraw(true);

                if (!dynamicGestures)
                {
                    m_rotY = y;
                    m_rotX = x;
                }
            }
            else
            {
                m_rotY = (double)rightHand.Position.Z - (double)leftHand.Position.Z;
                m_rotX = (double)rightHand.Position.Y - (double)leftHand.Position.Y;
            }
        }

        private void zoom()
        {
            var leftHand = m_closestSkeleton.Joints[JointType.HandLeft];
            var rightHand = m_closestSkeleton.Joints[JointType.HandRight];
            var leftElbow = m_closestSkeleton.Joints[JointType.ElbowLeft];
            var rightElbow = m_closestSkeleton.Joints[JointType.ElbowRight];
            var shoulderCenter = m_closestSkeleton.Joints[JointType.ShoulderCenter];

            bool left = false;
            bool right = false;

            if (leftHand.Position.Y > leftElbow.Position.Y && leftHand.Position.Z < shoulderCenter.Position.Z - 0.23 && !m_leftGrip)
                left = true;

            if (rightHand.Position.Y > rightElbow.Position.Y && rightHand.Position.Z < shoulderCenter.Position.Z - 0.23 && !m_rightGrip)
                right = true;

            if (!left || !right)
            {
                m_zoom = 0.0;
                m_gestures = 0;
                return;
            }

            m_gestures = 2;

            if (m_zoom != 0.0)
            {
                double x = Math.Abs((double)rightHand.Position.X - (double)leftHand.Position.X);
                double pom = 1.0;

                if (x > m_zoom + 0.009)
                {
                    pom = 0.975;
                }
                else
                    if (x < m_zoom - 0.009)
                    {
                        pom = 1.025;
                    }

                cam.Zoom(pom);
                scene.CurrentViewer.Redraw(true);

                if(!dynamicGestures)
                    m_zoom = x;
            }
            else
            {
                m_zoom = Math.Abs((double)rightHand.Position.X - (double)leftHand.Position.X);
            }
        }

        private void moveXYZ()
        {
            var leftHand = m_closestSkeleton.Joints[JointType.HandLeft];
            var rightHand = m_closestSkeleton.Joints[JointType.HandRight];
            var leftElbow = m_closestSkeleton.Joints[JointType.ElbowLeft];
            var rightElbow = m_closestSkeleton.Joints[JointType.ElbowRight];
            bool left = false;
            bool right = false;

            if (leftHand.Position.Y > leftElbow.Position.Y)
                left = true;

            if (rightHand.Position.Y > rightElbow.Position.Y)
                right = true;

            if ((left && m_leftGrip) && right)
            {
                m_lastPos.x = 0.0;
                m_lastPos.y = 0.0;
                m_lastPos.z = 0.0;
                m_gestures = 0;
                return;
            }

            if (right && m_rightGrip)
            {
                m_gestures = 1;
                if (m_lastPos.x != 0.0 && m_lastPos.y != 0.0 && m_lastPos.z != 0.0)
                {
                    double x = m_lastPos.x - (double)rightHand.Position.X;
                    double y = m_lastPos.y - (double)rightHand.Position.Y;
                    double z = m_lastPos.z - (double)rightHand.Position.Z;
                    int pom = 105;

                    if (dynamicGestures)
                    {
                        z = m_moveZStart - (double)rightHand.Position.Z;
                        pom = 10;
                    }

                    if (!left)
                    {
                        Move(x, esriCameraMovementType.esriCameraMoveLeft, esriCameraMovementType.esriCameraMoveRight, 105);
                        Move(y, esriCameraMovementType.esriCameraMoveDown, esriCameraMovementType.esriCameraMoveUp, 105);
                        Move(z, esriCameraMovementType.esriCameraMoveAway, esriCameraMovementType.esriCameraMoveToward, pom);
                    }
                    else
                    {
                        int angle = 0;
                        if (x > 0.0)
                        {
                            if (x > 0.005)
                                angle = 5;
                        }
                        else
                        {
                            if (x < -0.005)
                                angle = -5;
                        }

                        cam.HTurnAround(angle);
                    }

                    m_lastPos.x = (double)rightHand.Position.X;
                    m_lastPos.y = (double)rightHand.Position.Y;
                    m_lastPos.z = (double)rightHand.Position.Z;
                }
                else
                {
                    m_lastPos.x = (double)rightHand.Position.X;
                    m_lastPos.y = (double)rightHand.Position.Y;
                    m_lastPos.z = (double)rightHand.Position.Z;

                    m_moveZStart = (double)rightHand.Position.Z;
                }
            }
            else
            {
                m_lastPos.x = 0.0;
                m_lastPos.y = 0.0;
                m_lastPos.z = 0.0;
                m_gestures = 0;
            }
        }

        private void Move(double axis, esriCameraMovementType direction1, esriCameraMovementType direction2, int ratio)
        {
            if (axis > 0.0)
            {
                if (axis > 0.005)
                    cam.Move(direction1, 0.005 * (ratio * axis));
            }
            else
            {
                if (axis < -0.005)
                    cam.Move(direction2, 0.005 * (ratio * axis * -1));
            }
        }

        /// <summary>
        /// kontroluje trackovani potrebnych casti tela
        /// </summary>
        /// <param name="skeleton"></param>
        /// <returns></returns>
        private bool CheckJointsForTracking(Skeleton skeleton)
        {
            var head = skeleton.Joints[JointType.Head];
            var leftHand = skeleton.Joints[JointType.HandLeft];
            var rightHand = skeleton.Joints[JointType.HandRight];
            var leftElbow = skeleton.Joints[JointType.ElbowLeft];
            var rightElbow = skeleton.Joints[JointType.ElbowRight];
            var leftWrist = skeleton.Joints[JointType.WristLeft];
            var rightWrist = skeleton.Joints[JointType.WristRight];
            var shoulderCenter = skeleton.Joints[JointType.ShoulderCenter];

            if(head.TrackingState == JointTrackingState.NotTracked || 
               leftHand.TrackingState == JointTrackingState.NotTracked ||
               rightHand.TrackingState == JointTrackingState.NotTracked ||
               leftElbow.TrackingState == JointTrackingState.NotTracked ||
               rightElbow.TrackingState == JointTrackingState.NotTracked ||
               shoulderCenter.TrackingState == JointTrackingState.NotTracked ||
               leftWrist.TrackingState == JointTrackingState.NotTracked ||
               rightWrist.TrackingState == JointTrackingState.NotTracked)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        private void StartSensor()
        {
            foreach (var potentialSensor in KinectSensor.KinectSensors)
            {
                if (potentialSensor.Status == KinectStatus.Connected)
                {
                    this.sensor = potentialSensor;
                    break;
                }
            }

            if (this.sensor == null)
            {
                MessageBox.Show("zadny kinect v dosahu");
                //
                OnDeactivate();
                return;
            }

            try
            {
                //http://www.dotnetfunda.com/articles/show/2214/how-to-do-push-to-press-with-kinect-sdk-17
                this.sensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);
                this.sensor.SkeletonStream.Enable();//param);

                try
                {
                    this.sensor.DepthStream.Range = DepthRange.Default;
                    this.sensor.SkeletonStream.EnableTrackingInNearRange = false;
                }
                catch (InvalidOperationException)
                {
                    // Non Kinect for Windows devices do not support Near mode, so reset back to default mode.
                    this.sensor.DepthStream.Range = DepthRange.Default;
                    this.sensor.SkeletonStream.EnableTrackingInNearRange = false;
                }
            }
            catch (InvalidOperationException)
            {
                OnDeactivate();
                //return;
            }

            this.sensor.SkeletonFrameReady += this.SensorSkeletonFrameReady; // Add an event handler - metoda ktera se ma zavolat, kdyz vznikne novy frame se skeletony
            m_waveLeftHand.GestureRecognized += new EventHandler(m_waveLeftHand_GestureRecognized);
            m_waveRightHand.GestureRecognized += new EventHandler(m_waveRightHand_GestureRecognized);

            try
            {
                this.sensor.Start();
            }
            catch (IOException)
            {
                this.sensor = null;
                MessageBox.Show("nepodarilo se nastartovat kinect");
                //return;
                OnDeactivate();
            }  
        }

        void m_waveRightHand_GestureRecognized(object sender, EventArgs e)
        {
            if (dynamicGestures)
                dynamicGestures = false;
            else
                dynamicGestures = true;
        }

        void m_waveLeftHand_GestureRecognized(object sender, EventArgs e)
        {
            cam.SetDefaultsMBB(scene.Extent);
        }

        void interactionStream_InteractionFrameReady(object sender, InteractionFrameReadyEventArgs e)
        {
            using (InteractionFrame frame = e.OpenInteractionFrame())
            {
                if (frame != null)
                {
                    if (this.userInfos == null)
                    {
                        this.userInfos = new UserInfo[InteractionFrame.UserInfoArrayLength];
                    }
                    frame.CopyInteractionDataTo(this.userInfos);
                }
                else
                {
                    return;
                }
            }

            foreach (UserInfo userInfo in this.userInfos)
            {

                foreach (InteractionHandPointer handPointer in userInfo.HandPointers)
                {
                    string action = null;
                    bool gripped = false;

                    switch (handPointer.HandEventType)
                    {
                        case InteractionHandEventType.Grip:
                            action = "gripped";
                            gripped = true;
                            break;

                        case InteractionHandEventType.GripRelease:
                            action = "released";
                            gripped = false;
                            break;
                    }

                    if (action != null)
                    {
                        switch (handPointer.HandType)
                        {
                            case InteractionHandType.Left:
                                if (gripped)
                                    m_leftGrip = true;
                                else
                                    m_leftGrip = false;
                                break;

                            case InteractionHandType.Right:
                                if (gripped)
                                    m_rightGrip = true;
                                else
                                    m_rightGrip = false;
                                break;
                        }
                    }
                }
            }
        }

        protected override void OnKeyUp(ESRI.ArcGIS.Desktop.AddIns.Tool.KeyEventArgs arg)
        {
            cam.Apply();

            switch (arg.KeyCode)
            {
                case Keys.W:
                    cam.Move(esriCameraMovementType.esriCameraMoveAway, 0.05);
                    break;
                case Keys.A:
                    cam.Rotate(-5);
                    break;
                case Keys.S:
                    cam.Move(esriCameraMovementType.esriCameraMoveToward, 0.05);
                    break;
                case Keys.D:
                    cam.Rotate(5);
                    break;
                case Keys.Q:
                    cam.Move(esriCameraMovementType.esriCameraMoveUp, 0.05);
                    break;
                case Keys.E:
                    cam.Move(esriCameraMovementType.esriCameraMoveDown, 0.05);
                    break;
                case Keys.Y:
                    cam.Move(esriCameraMovementType.esriCameraMoveLeft, 0.05);
                    break;
                case Keys.X:
                    cam.Move(esriCameraMovementType.esriCameraMoveRight, 0.05);
                    break;
                case Keys.R:
                    cam.Zoom(1.1);
                    break;
                case Keys.T:
                    cam.Zoom(0.9);
                    break;
                case Keys.F:
                    cam.RollAngle = cam.RollAngle + 5;
                    break;
                case Keys.G:
                    cam.RollAngle = cam.RollAngle - 5;
                    break;
                case Keys.V:
                    cam.Inclination = cam.Inclination +10;
                    break;
                case Keys.B:
                    cam.Inclination = cam.Inclination - 10;
                    break;
            }
        }

        protected override void OnUpdate()
        {
            Enabled = ArcScene.Application != null;
        }

        private void ProcessDepthFrame(long skeletonTimestamp)
        {
            using (DepthImageFrame depthframe = this.sensor.DepthStream.OpenNextFrame(0))
            {
                if (depthframe != null)
                {
                    DepthImagePixel[] dPixels = new DepthImagePixel[depthframe.PixelDataLength];
                    depthframe.CopyDepthImagePixelDataTo(dPixels);
                    interactionStream.ProcessDepth(dPixels, skeletonTimestamp);
                }
            }
        }        
    }
}
