﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Kinect;
using ESRI.ArcGIS.ArcScene;
using ESRI.ArcGIS.Analyst3D;

namespace ArcSceneAddin2
{
    public class WaveSegment1 : GestureSegment
    {
        public WaveSegment1(JointType hand, JointType wrist, JointType elbow, JointType wrongHand, JointType wrongElbow)
        {
            m_hand = hand;
            m_wrist = wrist;
            m_elbow = elbow;
            m_elbowWrong = wrongElbow;
            m_handWrong = wrongHand;
        }

        public override bool Update(Skeleton skeleton, bool gripped)
        {
            if (skeleton.Joints[m_hand].Position.Y > skeleton.Joints[m_elbow].Position.Y &&
                skeleton.Joints[m_handWrong].Position.Y < skeleton.Joints[m_elbowWrong].Position.Y &&
                !gripped)
            {
                if (skeleton.Joints[m_hand].Position.X + 0.04 > skeleton.Joints[m_elbow].Position.X ||
                    skeleton.Joints[m_wrist].Position.X < skeleton.Joints[m_hand].Position.X)
                {
                    return true;
                }
            }

            return false;
        }
    }

    public class WaveSegment2 : GestureSegment
    {
        public WaveSegment2(JointType hand, JointType wrist, JointType elbow, JointType wrongHand, JointType wrongElbow)
        {
            m_hand = hand;
            m_wrist = wrist;
            m_elbow = elbow;
            m_elbowWrong = wrongElbow;
            m_handWrong = wrongHand;
        }

        public override bool Update(Skeleton skeleton, bool gripped)
        {
            if (skeleton.Joints[m_hand].Position.Y > skeleton.Joints[m_elbow].Position.Y &&
                skeleton.Joints[m_handWrong].Position.Y < skeleton.Joints[m_elbowWrong].Position.Y &&
                !gripped)
            {
                if (skeleton.Joints[m_hand].Position.X - 0.04 < skeleton.Joints[m_elbow].Position.X ||
                    skeleton.Joints[m_wrist].Position.X > skeleton.Joints[m_hand].Position.X)
                {
                    return true;
                }
            }

            return false;
        }
    }

    public class WaveGesture
    {
        GestureSegment[] m_segments;

        readonly int m_maxNumberOfFrames = 60;
        int m_currentSegment = 0;
        int m_frameCount = 0;

        public event EventHandler GestureRecognized;

        public WaveGesture(JointType hand, JointType wrist, JointType elbow, JointType wrongHand, JointType wrongElbow)
        {
            WaveSegment1 wave1 = new WaveSegment1(hand, wrist, elbow, wrongHand, wrongElbow);
            WaveSegment2 wave2 = new WaveSegment2(hand, wrist, elbow, wrongHand, wrongElbow);

            m_segments = new GestureSegment[]
            {
                wave1,
                wave2,
                wave1,
                wave2,
                wave1,
                wave2
            };
        }

        public void Update(Skeleton skeleton, bool gripped)
        {
            bool result = m_segments[m_currentSegment].Update(skeleton, gripped);

            if (result)
            {
                if (m_currentSegment + 1 < m_segments.Length)
                {
                    m_currentSegment++;
                    m_frameCount = 0;
                }
                else
                {
                    if (GestureRecognized != null)
                    {
                        GestureRecognized(this, new EventArgs());
                        Reset();
                    }
                }
            }
            else if (!result && m_frameCount == m_maxNumberOfFrames)
            {
                Reset();
            }
            else
            {
                m_frameCount++;
            }
        }

        public void Reset()
        {
            m_currentSegment = 0;
            m_frameCount = 0;
        }
    }
}
